require 'pretty_proxy'
require 'ballonizer'
require 'sequel'
require 'rack'
require 'addressable/uri'

# A PrettyProxy (that is a Rack::Proxy, that is a rack app) subclass that
# use an Ballonizer instance to allow the addition of speech bubbles (text
# ballons) over any (X)HTML page that is requested through it.
#
# The first three parameters are equal to the PrettyProxy ones, and the last
# two equal the Ballonizer ones. This class only deal in a special way with
# the *_asset_path_for_link and form_handler_url that are joined with the
# proxy site (scheme+host+port) to make a absolute url if the original
# values are relative.
#
# @author Henrique Becker 
class BallonizerProxy < PrettyProxy
  attr_reader :ballonizer

  def initialize(proxy_path, original_domain, original_paths, database, settings)
    super(proxy_path, original_domain, original_paths, true)
    @ballonizer = Ballonizer.new(database, settings)
  end

  def sugared_rewrite_response(triplet, requested_to_proxy_env, rewritten_env)
    status, headers, page = triplet
    original_url = Rack::Request.new(rewritten_env).url
    requested_to_proxy_url = Rack::Request.new(requested_to_proxy_env).url
    proxy_site = Addressable::URI.parse(requested_to_proxy_url).site

    absolute_settings = {}
    [ :css_asset_path_for_link,
      :js_asset_path_for_link,
      :form_handler_url].each do | conf_name |
      relative_uri = @ballonizer.settings[conf_name]
      next unless relative_uri 
      absolute_settings[conf_name] = Addressable::URI.parse(proxy_site)
                                                     .join(relative_uri).to_s
    end

    page = @ballonizer.ballonize_page(
      page, original_url, headers['content-type'], absolute_settings
    )

    # This makes the browser always request a new copy of the document (status
    # 200) and avoid 304 status without body to us ballonize (and the ballons
    # can be updated without the original page being updated so isn't
    # interesting to preserve the original cache control)
    Workarounds.make_uncacheable(headers)

    triplet = [status, headers, page]
  end

  module Workarounds
    # NOTE: THIS CHANGES THE ARGUMENT
    def self.make_uncacheable(headers)
      headers["Cache-Control"] = "no-cache, no-store, max-age=0, must-revalidate"
      headers["Pragma"] = "no-cache"
      headers["Expires"] = "Fri, 01 Jan 1990 00:00:00 GMT"
      headers
    end
  end

  private_constant :Workarounds
end

# A class to automate the creation of an rack app that offer multiple
# BallonizerProxys. The constructor only takes a hash (often parsed from
# a JSON config file) and create a rack app that can be accessed by #app.
# 
# The hash format is as follows:
#
# h['proxys_namespace'] has a String path (that start and end with slashes)
# all the proxy apps will be inside this namespace/'virtual directory'.
#
# h['ballonizer_proxys'] has an Array with the Hashes of the individual
# settings of each ballonizer proxy. The 'proxy_path', 'original_domain',
# 'original_paths' are as the PrettyProxy parameters with the same name.
# The 'database_uri' is equivalent to the database parameter of the
# Ballonizer constructor. These four are obrigatory.
# The rest of the options are as the ones described for the settings hash
# of the Ballonizer constructor. It's NOT RECOMMENDED to configure the
# :form_handler_url, :add_required_*, :add_js_for_edition, :*_asset_path_for_link,
# options, these options are automatically defined. The purpose of this
# class is mainly to handle this configuration for you, so doesn't make
# much sense to define these settings.
#
# Use 'rake example' to run the example. You can see the results in
# http://localhost:your_server_port/proxys/xkcd/pt-br/ and
# http://localhost:your_server_port/proxys/cardboad-crack/pt-br/.
# The configuration file is in example/settings_example.json
#
# This documentation is also avaliable in: http://www.omniref.com/ruby/gems/ballonizer_proxy
#
# @author Henrique Becker 
class BallonizerProxyGenerator
  module PrivateUtils
    def self.deep_freeze(o)
      if o.is_a? Hash
        o.each_pair { | k, v | deep_freeze(k); deep_freeze(v) }
      else
        if o.is_a? Array
          o.each { | e | deep_freeze(e) }
        end
      end

      o.freeze
    end

    def self.deep_clone(o)
      JSON.parse(JSON.generate(o))
    end
  end

  ASSET_PATH = '/assets/'

  private_constant :PrivateUtils, :ASSET_PATH

  def initialize(settings)
    @settings = PrivateUtils.deep_freeze(settings)
    @ballonizer_proxys = []
    @settings['ballonizer_proxys'].each do | bp_config |
      ballonizer_settings = {}
      Ballonizer::DEFAULT_SETTINGS.each_key do | key |
        value = bp_config[key.to_s]
        if value
          ballonizer_settings[key] = value
        end
      end
      ballonizer_settings = { 
        form_handler_url: bp_config['proxy_path'] + 'request_handler',
        add_required_css: true,
        add_required_js_libs_for_edition: true,
        add_js_for_edition: true,
        jquery_no_conflict: true,
        create_tables_if_none: true,
        # the paths for assets are made absolute in BallonizerProxy
        css_asset_path_for_link: ASSET_PATH,
        js_asset_path_for_link: ASSET_PATH
      }.merge(ballonizer_settings)
      @ballonizer_proxys << BallonizerProxy.new(
        @settings['proxys_namespace'] + bp_config['proxy_path'][1..-1],
        bp_config['original_domain'],
        bp_config['original_paths'],
        bp_config['database_uri'],
        ballonizer_settings
      )
    end

    proxys_namespace = @settings['proxys_namespace']
    ballonizer_proxys = @ballonizer_proxys

    @app = Rack::Builder.new do
      map ASSET_PATH do
        run Ballonizer.assets_app
      end
      ballonizer_proxys.each do | bp |
        map bp.proxy_path do
          run bp
        end

        map bp.ballonizer.settings[:form_handler_url] do
          run(lambda do | env |
            begin
              bp.ballonizer.process_submit(env)
              [200, {}, ['your changes have been stored successfully']]
            rescue Ballonizer::SubmitError => e
              [200, {}, [e.message]]
            end
          end)
        end
      end
    end
  end

  def app
    @app
  end

  # The proxy namespace can't be assets
#  def self.validate_settings(settings)
#  end
end

